/*
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree.
 */

#pragma once

#define TENSORPIPE_HAS_SHM_TRANSPORT 0
#define TENSORPIPE_HAS_IBV_TRANSPORT 0

#define TENSORPIPE_HAS_CMA_CHANNEL 0
