// ${generated_comment}

#include <ATen/RedispatchFunctions.h>
#include <ATen/Functions.h>

#include <ATen/core/dispatch/Dispatcher.h>
#include <ATen/core/op_registration/adaption.h>

namespace at {

namespace redispatch {
    ${function_redispatch_definitions}
} // namespace redispatch

} // namespace at
