#pragma once

#ifndef TORCH_INDUCTOR_CPP_WRAPPER
// All pure C++ headers for the C++ frontend.
#include <torch/all.h>
#endif

// Python bindings for the C++ frontend (includes Python.h).
#include <torch/python.h>
